import os
import traceback
from lantern_data_manager.utils.traccar import get_server_credentials
from lantern_data_manager.utils.logger import logger
from lantern_data_manager.traccar.api_controller import RestAPIController
from lantern_data_manager.traccar.exceptions import *

class TraccarController:

    def __init__(self,traccar_url,credentials):
        """Constructor of class

        Args:
            url (string): url of traccar server
            credentials (dict): {
                "user" : user in traccar,
                "password" : password
            }
        """
        # base_url, http_user, http_password = get_server_credentials(server_id=server_id)
        if "user" not in credentials or "password" not in credentials:
            raise BadParamsException("Not user or password in credentials")
        self.api = RestAPIController(base_url=traccar_url, http_user=credentials["user"], http_password=credentials["password"] )

    def get_devices(self):
        """Returns a list of all devices

            Returns:
                list<dict>: all devices
        """
        uri = 'api/devices'
        return self.api.http_get(uri)
    
    def get_device(self, device_id):
        """Gets a device by its uniqueId

        Args:
            device_id (string): device to get
        Returns:
            list<dict>: data of device found
        """
        # pull device from traccar
        uri = 'api/devices'
        params = {'uniqueId':device_id}
        try:
            return self.api.http_get(uri,params)
        except Exception as e:
            # traceback.print_exc()
            if "NullPointerException" in str(e):
                raise ObjectNotFoundException(device_id)
            else:
                raise e;

    def create_device(self, uniqueId, name, data={}, user=None):
        """Creates a new device
        Args:
            uniqueId (string): new device's id
            name (string): new device's name
            data (dict): data to agregate to the new device
            user (string): user to link the device
            
        Raises:
            DuplicateEntryException: if uniqueId is already in use

        Returns:
            list<dict>: the new device
        """
        # data = self.formatJSON(data)
        uri = 'api/devices'
        data["uniqueId"] = uniqueId
        data["name"] = name
        try:
            device = self.api.http_post(uri,data)
            if user:
                # logger.debug("adding link with user_id: {} and device_id: {}".format(user_id, device["id"]))
                self.link_device_to_user(device_id=uniqueId, user_id=user)
            return device
        except Exception as e:
            if "Duplicate entry" in str(e):
                raise DuplicateEntryException(uniqueId)
            else:
                raise e

    def update_device(self,id, uniqueId,name, data):
        """Updates a device by its id

        Args:
            id (int): device's id given by traccar
            uniqueId (string): unique device's id, chosen
            name (string): device's name, can be the current name or a new one
            data (dict): data to insert into the device

        Returns:
            dict: device data updated
        """
        data["name"] = name
        data["id"] = id
        data["uniqueId"] = uniqueId
        uri = 'api/devices/{}'.format(id)
        try:
            return self.api.http_put(uri,data)
        except Exception as e:
            if "NullPointerException" in str(e):
                raise ObjectNotFoundException(id)
            
    def delete_device(self,id):
        """Deletes a device by its id

        Args:
            id (int): device's id givem by traccar

        Returns:
            bool: true when device is deleted
        """
        uri = 'api/devices/{}'.format(id)
        return self.api.http_delete(uri)

    def list(self):
        uri = 'api/devices'
        return self.api.http_get(uri)
    
    def link_device_to_user(self, device_id, user_id):
        uri = 'api/permissions'
        data = { "userId": user_id, "deviceId": device_id } ## order in this dict is important
        try:
            self.api.http_post(uri, data)
        except Exception as e:
            pass
        return True

    def query_devices(self,uri,action,data={},params={}):
        """Performs a query into the specified uri

        Args:
            uri (string): url to perform query
            action (string): action to perform, it must be "get"|"post"|"put"|"delete"
            data (dict, optional): body of request, it applies to post and put. Defaults to {}.
            params (dict, optional): params of request, it applies to get. Defaults to {}.

        Raises:
            BadRequestException: if action is not "get"|"post"|"put"|"delete"

        Returns:
            dict: response from url
        """
        if action not in ["get","post","put","delete"]:
            raise BadRequestException("action must be get,post, put or delete")
        if action == "get":
            response = self.api.http_get(uri,params)
        elif action == "post":
            response = self.api.http_post(uri,data)
        elif action == "put":
            response = self.api.http_put(uri,data)
        elif action == "delete":
            response = self.api.http_delete(uri)
        return response
        


from lantern_data_manager.utils.logger import logger
import jwt

def extract_params(event, arguments,optionals=[]):
    """ Extract params from event (local or gateway) """
    data = event
    data = event["body"] if "body" in event and event["body"] else event
    data = event["query"] if "query" in event and event["query"] else data
    total_params = {}
    for name in arguments:
        if name not in data:
            logger.error("{} column is required in event with event: {}".format(name, event))
            raise Exception("{} column is required in event".format(name))
        total_params[name] = data[name]
    for name in optionals:
        if name in data:
            total_params[name] = data[name]
    return total_params


def get_token_attributes(event, names):
    """ decode and extract a list of names from decode jwt token (based on cognito) """
    try:
        authorization_header = event.get("headers", None).get("Authorization", None)
        authorization_header = authorization_header.replace("Bearer","")
        authorization_header = authorization_header.replace("bearer","")
        authorization_header = authorization_header.replace(" ", "")
        logger.info("authorization_header: {}".format(authorization_header))
        decoded = jwt.decode(authorization_header, verify=False)
        logger.info("decoded: {}".format(decoded))
        total_values = {}
        logger.info("names to extract: {}".format(names))
        for name in names:
            total_values[name] = decoded.get(name, None)
        return total_values
    except Exception as e:
        logger.error("Error decoding jwt and extracting values: {}".format(str(e)))


import os
import boto3
from moto import mock_sqs
# from lantern_data_manager.utils.logger import logger

# register variables in environment
def register_environment( variables ):
    """ register dictionary of variables in environment/memory variables """
    for key in variables.keys():
        os.environ[key] = variables[key]

# creates dynamodb table
def create_tables(instance,tables):
    for table in tables:
        # dynamodb = boto3.resource('dynamodb', region_name=table["region_name"], endpoint_url=table["endpoint_url"])
        params = {
            "TableName": table["table_name"],
            "KeySchema": [{'AttributeName': x["name"], 'KeyType': x["key_type"] } for x in table["keys"]],
            "AttributeDefinitions": [{'AttributeName': x["name"], 'AttributeType': x["type"]} for x in table["columns"]],
            "ProvisionedThroughput": { 'ReadCapacityUnits': 10, 'WriteCapacityUnits': 10 }
        }
        # adding indexes
        indexes = []
        if "indexes" in table:
            for index in table["indexes"]:
                indexes.append({
                    "IndexName": index["name"],
                    "KeySchema": [{  "AttributeName": x["name"], "KeyType": x["key_type"]} for x in index["keys"]],
                    "Projection": { "ProjectionType": "ALL" },
                    "ProvisionedThroughput": { 'ReadCapacityUnits': 10, 'WriteCapacityUnits': 10 }
                })
        if len(indexes):
            params["GlobalSecondaryIndexes"] = indexes
        
        # Creating table with defined parameters
        try:
            table = instance.create_table(**params)
        except Exception as e:
            logger.warning("warning creating table with message: {}".format(e))


from lantern_data_manager.utils.logger import logger

TYPE_FLOAT = "float"
TYPE_INT = "int"
TYPE_BOOL = "bool"
TYPE_STRING = "string"
TYPE_DURATION = "duration"
TYPE_TIME = "time"
TYPE_UINT = "uint"
TYPES = [
    TYPE_FLOAT,TYPE_INT,TYPE_BOOL,
    TYPE_STRING,TYPE_DURATION,TYPE_TIME,
    TYPE_UINT,
]

def make_query(bucket, start_range, device_ids, measurement=None, type=None, 
        stop_range=None, field=None, filters=None, groupers=None):
    """ Query InfluxDB using input parameters
        
    Arguments:
        bucket {str} -- bucket name to perform query
        start {str} -- flux start format string(-2d, -1m, -10w, etc)
        device_ids {List of str} -- device id (filter) to perform query
        
    Keyword Arguments:
        stop {str} -- flux stop format string(-2d, -1m, -10w, etc)
        measurement {str} -- '_measurement' value in influxdb (metric type) ex: watermeter
        field {str} -- '_field' value in influxdb (metric name), ex: counter, temperature, humidity, etc
        filters {list} -- List of {name: 'str', value: 'str'}, which is a list of filters to be added to query. 
        groupers {list} -- Which is the grouper to be used, with the form: {every: 'str', fn: 'str', type: 'str'(None)}, ex: {every: '10s', fn: 'sum', type: 'float'}, where type is optional and will cast data to that type.
    """

    query_str = ""

    def add_q(query_str, _str):
        return query_str + ' \n ' + _str
        
    # base bucket
    query_str = add_q(query_str, 'from(bucket: "{}")'.format(bucket))
        
    # ranges
    if not stop_range:
        query_str = add_q(query_str, "|> range(start: {})".format(start_range))
    else:
        query_str = add_q(query_str, "|> range(start: {}, stop: {})".format(start_range, stop_range))
    
    # device ids
    if device_ids and len(device_ids):
        _ff = '|> filter(fn: (r) =>'
        for idx, d_id in enumerate(device_ids):
            if idx != 0:
                _ff = _ff + " or" 
            _ff = _ff + ' r.device_id == "{value}"'.format(value=d_id)
        _ff = _ff + ')'
        query_str = add_q(query_str, _ff)
    
    # measurement
    if measurement:
        query_str = add_q(query_str, '|> filter(fn: (r) => r._measurement == "{}")'.format(measurement))
    
    # field
    if field:
        query_str = add_q(query_str, '|> filter(fn: (r) => r._field == "{}")'.format(field))
    
    # filters (OR, we have one or the other)
    if filters and len(filters):
        _ff = '|> filter(fn: (r) =>'
        for idx, f in enumerate(filters):
            if idx != 0:
                _ff = _ff + " or" 
            _ff = _ff + ' r.{name} == "{value}"'.format(name=f["name"], value=f["value"])
        _ff = _ff + ')'
        query_str = add_q(query_str, _ff)

    # casting global type if defined
    if type:
        query_str = add_q(query_str, _cast_type(type))
    logger.debug("query_str: {}".format(query_str))
    # List of queries to execute
    list_query_str = []
    if not groupers:
        """ no groupers, so we add the original query 1 time """
        list_query_str.append(query_str)
    if groupers and len(groupers):
        """ We have groupers, so we add 1 query for each grouper """
        for g in groupers:
            _type = ""
            if "type" in g:
                _type = _cast_type(g["type"])
            g_q_str = ''' {q} \n {type} \n |> aggregateWindow(every: {every}, fn: {fn}) \n |> yield(name: "{fn}")'''.format(
                type=_type, q=query_str, every=g["every"], fn=g["fn"]
            )
            list_query_str.append(g_q_str)
    logger.info("*** BEGİNNİNG *** queries to be executed")
    return list_query_str

def _cast_type(type):
    """ will return casting line based on input type """
    if type not in TYPES:
        raise Exception("{} not implemented, it should be one of: {}".format(type, TYPES))
    if type == TYPE_FLOAT:
        return '|> toFloat()'
    elif type == TYPE_INT:
        return '|> toInt()'
    elif type == TYPE_BOOL:
        return '|> toBool()'
    elif type == TYPE_STRING:
        return '|> toString()'
    elif type == TYPE_DURATION:
        return '|> toDuration()'
    elif type == TYPE_TIME:
        return '|> toTime()'
    elif type == TYPE_UINT:
        return '|> toUInt()'
    else:
        raise Exception("{} not implemented".format(type))


from elasticsearch import Elasticsearch
import os

class ElasticSearchController:
    """ 
    Class that wraps Elasticsearch API import
    """

    def __init__(self,host,stage):
        self.elasticInstance = Elasticsearch(hosts=[host])
        self.url = host
        self.stage = stage

    def query(self,index,query):
        """Performs a query in a given index in elastic search

        Args:
            index (string): name of index to perform query
            query (dict): a valid elastic search query

        Returns:
            dict: response from elastic
        """
        index = "{}-{}".format(index,self.stage)
        return self.elasticInstance.search(body=query, index=index)

    def add_data(self,index,id,document):
        """
        Creates new data in the index

        Args:
            index (string): name of index to perform query
            id (string): the id that the document will have
            document (dict): the json/dict to be saved in elastic

        Returns:
            dict: response from elastic
        """
        index = "{}-{}".format(index,self.stage)
        return self.elasticInstance.create(index=index,id=id,body=document)




from lantern_data_manager.traccar.traccar_controller import (
    TraccarController,
)
from lantern_data_manager.traccar.api_controller import (
    RestAPIController,
)
from lantern_data_manager.traccar.exceptions import (
    TraccarControllerException,
    BadRequestException,
    ObjectNotFoundException,
    ForbiddenAccessException,
    UserPermissionException,
    InvalidTokenException,
    BadParamsException,
    DuplicateEntryException,
)


import os
from lantern_data_manager.utils.logger import logger

def get_server_credentials(server_id="1"):
    """ Return Traccar Server URL, User, Pass
    based on server_id received in parameters, this should be a valid server. 
    Conditions:
        1. If server_id is None, we will return Main server creadentials which will exist always. (server with id=1)
        2. If server_id is defined, we will find a exact match with configuration, if no config is found a exception will be triggered.
        3. If server_id matches with configuration we will return that configuration.
        -- We will validate ALL 3 configuration variables are defined and have a valid value. 

    Args:
        server_id ([type]): [description]
    """
    logger.debug("get_server_credentials with server_id: {}".format(server_id))
    base_url_env_name = "TRACCAR_API_URL_{server_id}"
    base_user_env_name = "TRACCAR_API_HTTP_USER_{server_id}"
    base_pass_env_name = "TRACCAR_API_HTTP_PASS_{server_id}"

    if server_id is None:
        raise Exception("server_id is None")


    # getting full env variable names
    _url_env = base_url_env_name.format(server_id=server_id)
    _user_env = base_user_env_name.format(server_id=server_id)
    _pass_env = base_pass_env_name.format(server_id=server_id)
    
    # validating all those variables are defined in env
    _url_env_val = os.environ.get(_url_env, None)
    _user_env_val = os.environ.get(_user_env, None)
    _pass_env_val = os.environ.get(_pass_env, None)
    if not _url_env_val or not _user_env_val or not _pass_env_val:
        raise Exception("Configuration ERROR: At least one required ENV is not defined for selected Traccar instance. \
            Please check this variables in serverless: {}, {}, {}".format(_url_env, _user_env, _pass_env))
    
    return _url_env_val, _user_env_val, _pass_env_val



class TraccarControllerException(Exception):

    def __init__(self, info):
        """
        Args:
            info: Message to append to the error
        """
        self.info = info

    def __str__(self):
        return 'TraccarController error: {}'.format(self.info)


class BadRequestException(TraccarControllerException):
    def __init__(self, message):
        """
        Args:
            message:
        """
        super().__init__(info=message)


class ObjectNotFoundException(TraccarControllerException):

    def __init__(self, obj):
        """
        Args:
            obj:
            obj_type:
        """
        message = '[{} not found]'.format(obj)
        super().__init__(info=message)


class ForbiddenAccessException(TraccarControllerException):

    def __init__(self):
        message = '[Access is denied]: Wrong username or password'
        super().__init__(info=message)


class UserPermissionException(TraccarControllerException):

    def __init__(self):
        message = '[User has not enough permissions]'
        super().__init__(info=message)


class InvalidTokenException(TraccarControllerException):

    def __init__(self):
        message = '[Invalid user token]'
        super().__init__(info=message)

class BadParamsException(TraccarControllerException):
    
    def __init__(self,message):
        super().__init__(info=message)

class DuplicateEntryException(TraccarControllerException):
    def __init__(self,value):
        message = "Device already exist for uniqueId: {}".format(value)
        super().__init__(info=message)


from lantern_data_manager.utils.event import get_token_attributes

def changeFormat(response):
    for element in response:
        if 'Attributes' in element:
            for attribute in element['Attributes']:
                key = attribute.pop('Name')
                value = attribute.pop('Value')
                element[key] = value
            element.pop('Attributes')

def getTeamView(event):
    token = get_token_attributes(event, ["custom:view_team"])
    if event['body']:
        event["body"]["view_team"] = token.pop("custom:view_team")
    else:
        event["view_team"] = token.pop("custom:view_team")
    return event


import json
import os
from datetime import datetime
from io import StringIO

import requests
from influxdb_client import InfluxDBClient, Point, WritePrecision  # To handle InfluxDB
from influxdb_client.client.write_api import (
    SYNCHRONOUS,
)  # To handle InfluxDB synchonous writing
from lantern_data_manager.utils.logger import logger
from lantern_data_manager.utils.query import make_query
from lantern_data_manager.utils.time import datetime_to_str, to_nanoseconds
from requests.auth import HTTPBasicAuth

IS_LOCAL = os.environ.get("IS_LOCAL", False)


class InfluxController:
    def __init__(self, url=None, org=None, token=None):
        """Constructor of class

        Args:
            ulr (string): url to influx api, exampl: http://localhost:9999
            token (string, optional): token to influx, if absent, read from environ, otherwise fail.
            org (string, optional): organization to influx, if absent, read from environ, otherwise fail.
            user_name (string, optional): user_name to influx, if absent, read from environ, otherwise fail.
            password (string, optional): password to influx, if absent, read from environ, otherwise fail.
        """
        if not url:
            url = os.getenv("INFLUXDB_BASE_URL", False)
            if not url:
                raise Exception("url not defined")
        if IS_LOCAL:
            url = os.getenv("INFLUXDB_BASE_URL_PUBLIC", False)
        if not org:
            org = os.getenv("INFLUXDB_ORG", False)
            if not org:
                raise Exception("org not defined")

        if not token:
            token = os.getenv("INFLUXDB_TOKEN", False)
            if not token:
                raise Exception("token not defined")

        self.org = org
        self.client = self.__get_influx_connection(url, token, self.org)

    def __get_influx_connection(self, URL, TOKEN, ORG):
        return InfluxDBClient(url=URL, token=TOKEN, org=ORG)

    def add_telemetries(self, bucket, telemetries, org=None):
        """
        Adds telemetries into influx

        Args:
            bucket (string): bucket to put telemetries
            telemetries (list<dict>): points to write into influx, example:
            [{
                "measurement" : mesurement value,
                "tags" : {}, custom values
                "fields" : {}, custom values
                "time" : int
            }]
            org (string, optional): organization, if absent, read from environ . Defaults to None.
        """
        for data in telemetries:
            if not data["measurement"]:
                raise Exception("no 'measurement' column found")
            if not data["tags"]:
                raise Exception("no 'tags' column found")
            if not data["fields"]:
                raise Exception("no 'fields' column found")
            if not data["time"]:
                raise Exception("no 'time' column found")

        write_client = self.client.write_api(write_options=SYNCHRONOUS)
        if not org:
            org = self.org
        points = [p.to_line_protocol() for p in self._get_points(telemetries)]
        return write_client.write(bucket, org, points)

    def _get_points(self, telemetries):
        _points = self._format_telemetries(telemetries)
        points = []
        for _point in _points:
            _point["timestamp"] = to_nanoseconds(round(_point["timestamp"]))
            point = Point(_point["measurement"])
            # adding tags
            try:
                _point["field"]["value"] = float(_point["field"]["value"])
            except:
                pass
            for tag in _point["tags"]:
                point.tag(tag["key"], tag["value"])
            point.field(_point["field"]["key"], _point["field"]["value"])
            point.time(_point["timestamp"])
            points.append(point)
        return points

    def _format_telemetries(self, telemetries):
        points = []
        for telemetry in telemetries:
            for field in telemetry["fields"]:
                point = {}
                point["measurement"] = telemetry["measurement"]
                point["tags"] = [
                    {"key": key, "value": value}
                    for key, value in telemetry["tags"].items()
                ]
                point["field"] = {"key": field, "value": telemetry["fields"][field]}
                point["timestamp"] = telemetry["time"]
                points.append(point)
        return points

    def call_query_builder(
        self,
        bucket,
        start_range,
        device_ids,
        measurement=None,
        type_=None,
        stop_range=None,
        field=None,
        filters=None,
        groupers=None,
    ):
        """Creates a query list based on the params

        Args:
            bucket (string): bucket
            start_range (string): a start point to query data: -1m,-1h,-1d, etc
            device_ids (list): list of devices to query
            measurement (string, optional): '_measurement' value in influxdb (metric type) ex: watermeter. Defaults to None.
            stop_range (string, optional): an stop point to query data, similar to start_range. Defaults to None.
            field (string, optional): '_field' value in influxdb (metric name), ex: counter, temperature, humidity, etc. Defaults to None.
            filters (list<dict>, optional): List of {name: 'str', value: 'str'}, which is a list of filters to be added to query. Defaults to None.
            groupers (list<dict>, optional): Which is the grouper to be used, with the form: {every: 'str', fn: 'str', type: 'str'(None)}, ex: {every: '10s', fn: 'sum', type: 'float'}, where type is optional and will cast data to that type.

        Returns:
            list<string>: list with queries, it's a query per grouper
        """
        return make_query(
            bucket,
            start_range,
            device_ids,
            measurement,
            type_,
            stop_range,
            field,
            filters,
            groupers,
        )

    def query(self, query_string):
        """Performs a query into influx

        Args:
            query_string (string): query to perform in influx, for an easier use,
                                  call call_query_builder to generate a query

        Returns:
            list<dict>: response from influx
        """
        query_api = self.client.query_api()
        tables = query_api.query(query_string)
        return self.parse_data(tables)

    def parse_data(self, data):
        items = []
        for table in data:
            for record in table.records:
                items.append(record.values)
        return items


""" Usage example
        gateway_trace = {
            "mapping" : (self.timestamp)*1000,
            "virtual_node" : "virtual" + self.virtual_node,
            "time_interval": self.time_interval,
            "gateway_mac": self.gateway_mac,
            "seen_beacons": len(self.data),
            "gateways": len(self.data)
        }
        logger.info("Saving gateways telemetries into InfluxDB for: {}".format(gateway_trace))
        influx_manager = InfluxController('http://telemetry.lantern.rocks:9999','my-token','lantern',"{}_ble_buffer".format(self.team))
        influx_manager.write_data("Trilateration","gateways",[gateway_trace],"mapping")
        logger.info("TRACE GATEWAY {}".format(self.gateway_mac))
"""



"""
Lantern Data Manager
"""
from setuptools import find_packages, setup

PACKAGE_VERSION = "1.0.10"

DEPENDENCIES = [
    "elasticsearch==7.9.1",
    "PyJWT==1.7.1",
    "influxdb-client==1.14.0",
]

setup(
    name="lantern-data-manager",
    version=PACKAGE_VERSION,
    url="https://gitlab.com/lantern-tech/lantern-data-manager.git",
    license="BSD",
    author="Lantern Technologies",
    author_email="support@lantern.tech",
    description="Lantern Data Manager Library",
    long_description=__doc__,
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    zip_safe=False,
    platforms="any",
    install_requires=DEPENDENCIES,
    classifiers=[
        # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        # 'Development Status :: 1 - Planning',
        # 'Development Status :: 2 - Pre-Alpha',
        # 'Development Status :: 3 - Alpha',
        # 'Development Status :: 4 - Beta',
        "Development Status :: 5 - Production/Stable",
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX",
        "Operating System :: MacOS",
        "Operating System :: Unix",
        "Operating System :: Microsoft :: Windows",
        # 'Programming Language :: Python',
        # 'Programming Language :: Python :: 2',
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



import json, boto3
from lantern_data_manager.utils.logger import logger
from lantern_data_manager.utils.time import ts_to_iso8601_str

class TimestreamController():
    
    def __init__(self, profile_name=None):
        self.client = self._get_client(profile_name, 'timestream-query')
        self.EQUIVALENT_TYPES = {
            'VARCHAR': str,
            'DOUBLE': float,
            'BOOLEAN': self._format_bool,
            'TIMESTAMP': str
        }
                
    def _format_bool(self, bool_str):
        return eval('{}'.format(bool_str.replace('true', 'True').replace('false', 'False')))
                
    def _get_client(self, profile_name, client_operation_name):
        if profile_name:
            try:
                logger.info('Getting client for profile: {}'.format(profile_name))
                session = boto3.Session(profile_name=profile_name)
                return session.client(client_operation_name)
            except Exception as e:
                logger.error('Could not get connection with the specified profile: {}, error: {}'.format(profile_name, e))
                logger.info('Getting client with default profile.')
                return boto3.client(client_operation_name)
        else:
            logger.info('Getting client with default profile.')
            return boto3.client(client_operation_name)
        
    def _normalize_dates(self, from_date, to_date):
        try:
            from_date = "from_iso8601_timestamp('{}')".format(ts_to_iso8601_str(int(from_date)))
        except Exception as e:
            from_date = 'ago({})'.format(from_date)
            
        if to_date:
            try:
                to_date = "from_iso8601_timestamp('{}')".format(ts_to_iso8601_str(int(to_date)))
            except Exception as e:
                to_date = 'ago({})'.format(to_date)
        return from_date, to_date
        
    def _build_query_str(self, device_ids, database, from_date, to_date, measurement, grouper, variable):
        query_str = 'SELECT'
        group_by = False
        if grouper:
            if grouper['functions'] and grouper['functions'][0]:
                group_by = True
                query_str = '{} device_id,'.format(query_str)
                for fn in grouper['functions']:
                    query_str = '{} {}(measure_value::double) AS {},'.format(query_str, fn['fn'], fn['name'])
                query_str = '{} BIN(time, {}) AS time'.format(query_str, grouper['every'])
        if not group_by:
            query_str = '{} *'.format(query_str)
        query_str = '{} FROM {}.{} WHERE time'.format(query_str, database, measurement)
        from_date, to_date = self._normalize_dates(from_date, to_date)
        if to_date:
            query_str = '{} BETWEEN {} AND {}'.format(query_str, from_date, to_date)
        else:
            query_str = '{} > {}'.format(query_str, from_date)
        if variable:
            query_str = "{} AND measure_name = '{}'".format(query_str, variable)
            
        if device_ids:
            query_str = '{} AND device_id in {}'.format(query_str, str(tuple(device_ids)).replace(',)',')'))
        if group_by:
            query_str = '{} GROUP BY device_id, BIN(time, {})'.format(query_str, grouper['every'])
        query_str = '{} ORDER BY time ASC'.format(query_str)
        return query_str
        
    def _get_rows(self, query_str, rows):
        paginator = self.client.get_paginator('query')
        response_iterator = paginator.paginate(QueryString=query_str, PaginationConfig={'MaxItems': rows, 'PageSize': rows})
        rows = []
        columns = []
        for response in response_iterator:
            if not columns:
                columns = response['ColumnInfo']
            for row in response['Rows']:
                _row = {}
                for idx, value in enumerate(row['Data']):
                    if 'NullValue' not in value:
                        column = columns[idx]
                        key = column['Name'] if 'measure_value' not in column['Name'] else 'value'
                        val = self.EQUIVALENT_TYPES[column['Type']['ScalarType']](value['ScalarValue'])
                        _row[key] = val
                rows.append(_row)
        return rows
        
    def query(self, database, device_ids, from_date, rows, measurement, to_date=None, grouper={}, variable=None):
        query_str = self._build_query_str(device_ids, database, from_date, to_date, measurement, grouper, variable)
        rows = self._get_rows(query_str, rows)
        return rows


import requests
import json

class RestAPIController(object):

    def __init__(self, base_url, header_token_name=None, api_token=None, http_user=None, http_password=None):
        """ Initialize RestAPI
        
        Arguments:
            base_url {[type]} -- [description]
        
        Keyword Arguments:
            header_token_name {str} -- [header name to send api key if defined] (default: {None})
            api_token {str} -- [api key if required] (default: {None})
            http_user {str} -- [basic Auth user if required] (default: {None})
            http_password {str} -- [basic pass if required] (default: {None})
            if user/pass provided and apikey and header, api key will be used
        """
        self.base_url = base_url
        self.header_token_name = header_token_name
        self.api_token = api_token
        self.http_user = http_user
        self.http_password = http_password
        self.code_responses = [200,201,204]
        # self.session = requests.Session()


    def http_get(self, uri, params={}):
        """ HTTP REST GET, using params defined in params
        
        Arguments:
            uri {str} -- uri to add to base_url
        
        Keyword Arguments:
            params {dict} -- parameters to request (default: {{}} --- can be empty)
        
        Returns:
            [dict] -- request response content
        """
        full_path = '{}{}'.format(self.base_url,uri)
        if params:
            full_path = '{}?'.format(full_path)
            for index, key in enumerate(params.keys()):
                full_path = '{}{}={}'.format(full_path,key,params[key])
                if (index+1) != len(params.keys()):
                    full_path = '{}&'.format(full_path)
        response = requests.get(full_path, auth=(self.http_user,self.http_password))
        if response.status_code not in self.code_responses:
            if "message" in response.text:
                raise Exception(response.json()["message"])
            else:
                raise Exception(response.text)
        return response.json()

    
    def http_post(self, uri, data):
        """ HTTP REST POST, using data defined in data
        
        Arguments:
            uri {str} -- uri to add to base_url
            data {dict} -- data to be send in POST request
        
        Returns:
            [dict] -- request response content, raise an error if something wrong
        """
        full_path = '{}{}'.format(self.base_url,uri)
        headers = {'Content-Type': 'application/json'}
        response = requests.post(full_path, data=json.dumps(data), headers=headers, auth=(self.http_user,self.http_password))
        if response.status_code not in self.code_responses:
            if "message" in response.text:
                print(response.json())
                raise Exception(response.json()["message"])
            else:
                raise Exception(response.text)
        return response.json()

    def http_put(self, uri, data):
        """ HTTP REST PUT, using data defined in data
        
        Arguments:
            uri {str} -- uri to add to base_url
            data {dict} -- data to be send in PUT request
        
        Returns:
            [dict] -- request response content
        """
        full_path = '{}{}'.format(self.base_url,uri)
        headers = {'Content-Type': 'application/json'}
        response = requests.put(full_path, data=json.dumps(data), headers=headers, auth=(self.http_user,self.http_password))
        if response.status_code not in self.code_responses:
            if "message" in response.text:
                raise Exception(response.json()["message"])
            else:
                raise Exception(response.text)
        return response.json() if response.text != "" else data

    def http_delete(self, uri):
        """ HTTP REST DELETE, using device_id in uri
        
        Arguments:
            uri {str} -- uri to add to base_url
        
        Returns:
            [str] -- request response content
        """
        full_path = '{}{}'.format(self.base_url,uri)
        response = requests.delete(full_path, auth=(self.http_user,self.http_password))
        if response.status_code not in self.code_responses:
            if "message" in response.text:
                raise Exception(response.json()["message"])
            else:
                raise Exception(response.text)
        return response.json() if response.text != "" else True